#!/usr/bin/env python
"""The PROSPECT leaf optical properties model
Versions 5 and D

"""
import numpy as np
from scipy.special import expi

from prosail import spectral_lib

def run_prospect(n, cab, car,  cbrown, cw, cm, ant=0.0, 
                 prospect_version="D",  
                 nr=None, kab=None, kcar=None, kbrown=None, kw=None, 
                 km=None, kant=None, alpha=40.):
    """The PROSPECT model, versions 5 and D"""
    
    if prospect_version == "5":
        # Call the original PROSPECT-5. In case the user has supplied 
        # spectra, use them.
        wv, refl, trans = prospect_d (n, cab, car, cbrown, cw, cm, 0.0,
                    spectral_lib.prospect5.nr if nr is None else nr,
                    spectral_lib.prospect5.kab if kab is None else kab,
                    spectral_lib.prospect5.kcar if kcar is None else kcar,
                    spectral_lib.prospect5.kbrown \
                        if kbrown is None else kbrown, 
                    spectral_lib.prospect5.kw if kw is None else kw,
                    spectral_lib.prospect5.km if km is None else km,
                    np.zeros_like(spectral_lib.prospect5.km), 
                    alpha=alpha)
    elif prospect_version.upper() == "D":
        wv, refl, trans = prospect_d (n, cab, car, cbrown, cw, cm, ant,
                    spectral_lib.prospectd.nr if nr is None else nr,
                    spectral_lib.prospectd.kab if kab is None else kab,
                    spectral_lib.prospectd.kcar if kcar is None else kcar,
                    spectral_lib.prospectd.kbrown \
                        if kbrown is None else kbrown,
                    spectral_lib.prospectd.kw if kw is None else kw,
                    spectral_lib.prospectd.km if km is None else km,
                    spectral_lib.prospectd.kant if kant is None else kant, 
                    alpha=alpha)
    else:
        raise ValueError("prospect_version can only be 5 or D!")

    return wv, refl, trans


def calctav ( alpha, nr):
    #***********************************************************************
    #calctav
    # ***********************************************************************
    # Stern F. (1964), Transmission of isotropic radiation across an
    # interface between two dielectrics, Appl. Opt., 3(1):111-113.
    # Allen W.A. (1973), Transmission of isotropic light across a
    # dielectric surface in two and three dimensions, J. Opt. Soc. Am.,
    # 63(6):664-666.
    # ***********************************************************************
    

    #rd  = pi/180 np.deg2rad
    n2  = nr*nr
    npx  = n2+1
    nm  = n2-1
    a   = (nr+1)*(nr+1)/2.
    k   = -(n2-1)*(n2-1)/4.
    sa  = np.sin(np.deg2rad(alpha))
    
    if alpha != 90:
        b1 = np.sqrt((sa*sa-npx/2)*(sa*sa-npx/2)+k)
    else:
        b1 = 0.
    b2 = sa*sa - npx/2
    b   = b1-b2;
    b3  = b**3;
    a3  = a**3;
    ts  = (k**2/(6*b3)+k/b-b/2)-(k**2./(6*a3)+k/a-a/2)

    tp1 = -2*n2*(b-a)/(npx**2)
    tp2 = -2*n2*npx*np.log(b/a)/(nm**2)
    tp3 = n2*(1/b-1/a)/2
    tp4 = 16*n2**2*(n2**2+1)*np.log((2*npx*b-nm**2)/(2*npx*a-nm**2))/(npx**3*nm**2)
    tp5 = 16*n2**3*(1./(2*npx*b-nm**2)-1/(2*npx*a-nm**2))/(npx**3)
    tp  = tp1+tp2+tp3+tp4+tp5
    tav = (ts+tp)/(2*sa**2)

    return tav


def refl_trans_one_layer (alpha, nr, tau):
    # ***********************************************************************
    # reflectance and transmittance of one layer
    # ***********************************************************************
    # Allen W.A., Gausman H.W., Richardson A.J., Thomas J.R. (1969),
    # Interaction of isotropic ligth with a compact plant leaf, J. Opt.
    # Soc. Am., 59(10):1376-1379.
    # ***********************************************************************
    # reflectivity and transmissivity at the interface
    #-------------------------------------------------   
    talf = calctav (alpha,nr)
    ralf = 1.0-talf
    t12 = calctav (90,nr)
    r12 = 1. - t12
    t21 = t12/(nr*nr)
    r21 = 1-t21

    # top surface side
    denom = 1. - r21*r21*tau*tau
    Ta = talf*tau*t21/denom
    Ra = ralf + r21*tau*Ta

    # bottom surface side
    t = t12*tau*t21/denom
    r = r12+r21*tau*t
    
    return r, t, Ra, Ta, denom

def prospect_d (N, cab, car, cbrown, cw, cm, ant,
            nr, kab, kcar, kbrown, kw, km, kant,
            alpha=40.):

    lambdas = np.arange(400, 2501) # wavelengths
    n_lambdas = len(lambdas)
    n_elems_list = [len(spectrum)  for spectrum in 
                [nr, kab, kcar, kbrown, kw, km, kant]]
    if not all(n_elems == n_lambdas for n_elems in n_elems_list):
        raise ValueError("Leaf spectra don't have the right shape!")
    
    kall = (cab*kab + car*kcar + ant*kant + cbrown*kbrown +
            cw*kw + cm*km)/N
    j = kall > 0
    t1 = (1-kall)*np.exp(-kall)
    t2 = kall**2*(-expi(-kall))
    tau = np.ones_like(t1)
    tau[j] = t1[j] + t2[j]

    r, t, Ra, Ta, denom = refl_trans_one_layer (alpha, nr, tau)

    # ***********************************************************************
    # reflectance and transmittance of N layers
    # Stokes equations to compute properties of next N-1 layers (N real)
    # Normal case
    # ***********************************************************************
    # Stokes G.G. (1862), On the intensity of the light reflected from
    # or transmitted through a pile of plates, Proc. Roy. Soc. Lond.,
    # 11:545-556.
    # ***********************************************************************
    D       = np.sqrt((1+r+t)*(1+r-t)*(1.-r+t)*(1.-r-t))
    rq      = r*r
    tq      = t*t
    a       = (1+rq-tq+D)/(2*r)
    b       = (1-rq+tq+D)/(2*t)

    bNm1    = np.power(b, N-1)
    bN2     = bNm1*bNm1
    a2      = a*a
    denom   = a2*bN2-1
    Rsub    = a*(bN2-1)/denom
    Tsub    = bNm1*(a2-1)/denom

    # Case of zero absorption
    j       = r+t >= 1.
    Tsub[j] = t[j]/(t[j]+(1-t[j])*(N-1))
    Rsub[j] = 1-Tsub[j]

    # Reflectance and transmittance of the leaf: combine top layer with next N-1 layers
    denom   = 1-Rsub*r
    tran    = Ta*Tsub/denom
    refl    = Ra+Ta*Rsub*t/denom

    return lambdas, refl, tran

###if __name__ == "__main__":
    
    ###k_cab = prosail.spectral_libs.k_cab
    ###k_w = prosail.spectral_libs.k_cw
    ###k_m = prosail.spectral_libs.k_cm
    ###k_car = prosail.spectral_libs.k_car
    ###k_brown = prosail.spectral_libs.k_brown
    ###nr = prosail.spectral_libs.refractive

    ###wv, r, t = prospect_d (2.1, 60., 10., 0.1, 0.013, 0.016, 0,
            ###nr, k_cab, k_car, k_brown, k_w, k_m, k_m*0.,
            ###alpha=40.)
    
    ###rt = prosail.prospect_5b(2.1, 60., 10., 0.1, 0.013, 0.016)
    ###plt.plot(wv, r-rt[:,0], '--')
    ###plt.plot(wv, t-rt[:,1], '--')

    ###wv, r, t = prospect_d (2.1, 10., 10., 0.1, 0.013, 0.016, 0,
            ###nr, k_cab, k_car, k_brown, k_w, k_m, k_m*0.,
            ###alpha=40.)
####    plt.plot(wv, r)
    ###rt = prosail.prospect_5b(2.1, 10., 10., 0.1, 0.013, 0.016)
    ###plt.plot(wv, r-rt[:,0], '-')
    ###plt.plot(wv, t-rt[:,1], '-')



#!/usr/bin/env python
import numpy as np

from prosail import spectral_lib

from .prospect_d import run_prospect
from .FourSAIL import foursail

def run_prosail(n, cab, car,  cbrown, cw, cm, lai, lidfa, hspot,
                tts, tto, psi, ant=0.0, alpha=40., prospect_version="5", 
                typelidf=2, lidfb=0., factor="SDR",
                rsoil0=None, rsoil=None, psoil=None,
                soil_spectrum1=None, soil_spectrum2=None):
    """Run the PROSPECT_5B and SAILh radiative transfer models. The soil
    model is a linear mixture model, where two spectra are combined together as

         rho_soil = rsoil*(psoil*soil_spectrum1+(1-psoil)*soil_spectrum2)
    By default, ``soil_spectrum1`` is a dry soil, and ``soil_spectrum2`` is a
    wet soil, so in that case, ``psoil`` is a surface soil moisture parameter.
    ``rsoil`` is a  soil brightness term. You can provide one or the two
    soil spectra if you want.  The soil spectra must be defined
    between 400 and 2500 nm with 1nm spacing.

    Parameters
    ----------
    n: float
        Leaf layers
    cab: float
        leaf chlorophyll concentration
    car: float
        leaf carotenoid concentration
    cbrown: float
        senescent pigment
    cw: float
        equivalent leaf water
    cm: float
        leaf dry matter
    lai: float
        leaf area index
    lidfa: float
        a parameter for leaf angle distribution. If ``typliedf``=2, average
        leaf inclination angle.
    tts: float
        Solar zenith angle
    tto: float
        Sensor zenith angle
    psi: float
        Relative sensor-solar azimuth angle ( saa - vaa )
    ant: float
        leaf anthocyanin concentration (default set to 0)
    alpha: float
        The alpha angle (in degrees) used in the surface scattering
        calculations. By default it's set to 40 degrees.
    prospect_version: str
        Which PROSPECT version to use. We have "5" and "D"
    typelidf: int, optional
        The type of leaf angle distribution function to use. By default, is set
        to 2.
    lidfb: float, optional
        b parameter for leaf angle distribution. If ``typelidf``=2, ignored
    factor: str, optional
        What reflectance factor to return:
        * "SDR": directional reflectance factor (default)
        * "BHR": bi-hemispherical r. f.
        * "DHR": Directional-Hemispherical r. f. (directional illumination)
        * "HDR": Hemispherical-Directional r. f. (directional view)
        * "ALL": All of them
    rsoil0: float, optional
        The soil reflectance spectrum
    rsoil: float, optional
        Soil scalar 1 (brightness)
    psoil: float, optional
        Soil scalar 2 (moisture)
    soil_spectrum1: 2101-element array
        First component of the soil spectrum
    soil_spectrum2: 2101-element array
        Second component of the soil spectrum
    Returns
    --------
    A reflectance factor between 400 and 2500 nm


    """

    factor = factor.upper()
    if factor not in ["SDR", "BHR", "DHR", "HDR", "ALL"]:
        raise ValueError(
            "'factor' must be one of SDR, BHR, DHR, HDR or ALL"
        )
    if soil_spectrum1 is not None:
        assert (len(soil_spectrum1) == 2101)
    else:
        soil_spectrum1 = spectral_lib.soil.rsoil1

    if soil_spectrum2 is not None:
        assert (len(soil_spectrum1) == 2101)
    else:
        soil_spectrum2 = spectral_lib.soil.rsoil2

    if rsoil0 is None:
        if (rsoil is None) or (psoil is None):
            raise ValueError(
                "If rsoil0 isn't define, then rsoil and psoil"
                " need to be defined!"
            )
        rsoil0 = rsoil * (
        psoil * soil_spectrum1 + (1. - psoil) * soil_spectrum2)

    wv, refl, trans = run_prospect (n, cab, car,  cbrown, cw, cm, ant=ant, 
                 prospect_version=prospect_version, alpha=alpha)
    
    [tss, too, tsstoo, rdd, tdd, rsd, tsd, rdo, tdo,
         rso, rsos, rsod, rddt, rsdt, rdot, rsodt, rsost, rsot,
         gammasdf, gammasdb, gammaso] = foursail (refl, trans,  
                                                  lidfa, lidfb, typelidf, 
                                                  lai, hspot, 
                                                  tts, tto, psi, rsoil0)

    if factor == "SDR":
        return rsot
    elif factor == "BHR":
        return rddt
    elif factor == "DHR":
        return rsdt
    elif factor == "HDR":
        return rdot
    elif factor == "ALL":
        return [rsot, rddt, rsdt, rdot]


def run_sail(refl, trans, lai, lidfa, hspot, tts, tto, psi,
             typelidf=2, lidfb=0., factor="SDR",
             rsoil0=None, rsoil=None, psoil=None,
             soil_spectrum1=None, soil_spectrum2=None):
    """Run the SAILh radiative transfer model. The soil model is a linear
    mixture model, where two spectra are combined together as

         rho_soil = rsoil*(psoil*soil_spectrum1+(1-psoil)*soil_spectrum2)

    By default, ``soil_spectrum1`` is a dry soil, and ``soil_spectrum2`` is a
    wet soil, so in that case, ``psoil`` is a surface soil moisture parameter.
    ``rsoil`` is a  soil brightness term. You can provide one or the two
    soil spectra if you want. The soil spectra, and leaf spectra must be defined
    between 400 and 2500 nm with 1nm spacing.

    Parameters
    ----------
    refl: 2101-element array
        Leaf reflectance
    trans: 2101-element array
        leaf transmittance
    lai: float
        leaf area index
    lidfa: float
        a parameter for leaf angle distribution. If ``typliedf``=2, average
        leaf inclination angle.
    hspot: float
        The hotspot parameter
    tts: float
        Solar zenith angle
    tto: float
        Sensor zenith angle
    psi: float
        Relative sensor-solar azimuth angle ( saa - vaa )
    typelidf: int, optional
        The type of leaf angle distribution function to use. By default, is set
        to 2.
    lidfb: float, optional
        b parameter for leaf angle distribution. If ``typelidf``=2, ignored
    factor: str, optional
        What reflectance factor to return:
        * "SDR": directional reflectance factor (default)
        * "BHR": bi-hemispherical r. f.
        * "DHR": Directional-Hemispherical r. f. (directional illumination)
        * "HDR": Hemispherical-Directional r. f. (directional view)
        * "ALL": All of them
    rsoil0: float, optional
        The soil reflectance spectrum
    rsoil: float, optional
        Soil scalar 1 (brightness)
    psoil: float, optional
        Soil scalar 2 (moisture)
    soil_spectrum1: 2101-element array
        First component of the soil spectrum
    soil_spectrum2: 2101-element array
        Second component of the soil spectrum

    Returns
    --------
    Directional surface reflectance between 400 and 2500 nm


    """

    factor = factor.upper()
    if factor not in ["SDR", "BHR", "DHR", "HDR", "ALL"]:
        raise ValueError(
            "'factor' must be one of SDR, BHR, DHR, HDR or ALL"
        )
    if soil_spectrum1 is not None:
        assert (len(soil_spectrum1) == 2101)
    else:
        soil_spectrum1 = spectral_lib.soil.rsoil1

    if soil_spectrum2 is not None:
        assert (len(soil_spectrum1) == 2101)
    else:
        soil_spectrum2 = spectral_lib.soil.rsoil2

    if rsoil0 is None:
        if (rsoil is None) or (psoil is None):
            raise ValueError(
                "If rsoil0 isn't define, then rsoil and psoil"
                " need to be defined!"
            )
        rsoil0 = rsoil * (
        psoil * soil_spectrum1 + (1. - psoil) * soil_spectrum2)

    
    [tss, too, tsstoo, rdd, tdd, rsd, tsd, rdo, tdo,
         rso, rsos, rsod, rddt, rsdt, rdot, rsodt, rsost, rsot,
         gammasdf, gammasdb, gammaso] = foursail (refl, trans,  
                                                  lidfa, lidfb, typelidf, 
                                                  lai, hspot, 
                                                  tts, tto, psi, rsoil0)

    if factor == "SDR":
        return rsot
    elif factor == "BHR":
        return rddt
    elif factor == "DHR":
        return rsdt
    elif factor == "HDR":
        return rdot
    elif factor == "ALL":
        return [rsot, rddt, rsdt, rdot]


def run_thermal_sail(lam,  
                     tveg, tsoil, tveg_sunlit, tsoil_sunlit, t_atm, 
                     lai, lidfa, hspot,  
                     tts, tto, psi, rsoil=None,
                     refl=None, emv=None, ems=None,
                     typelidf=2, lidfb=0):
    c1 = 3.741856E-16
    c2 = 14388.0
    # Calculate the thermal emission from the different
    # components using Planck's Law
    top = (1.0e-6)*c1*(lam*1e-6)**(-5.)
    Hc = top / ( np.exp ( c2/(lam*tveg))-1.)         # Shade leaves
    Hh = top / ( np.exp ( c2/(lam*tveg_sunlit))-1.)  # Sunlit leaves
    Hd = top / ( np.exp ( c2/(lam*tsoil))-1.)        # shade soil 
    Hs = top / ( np.exp ( c2/(lam*tsoil_sunlit))-1.) # Sunlit soil
    Hsky = top / ( np.exp ( c2/(lam*t_atm))-1.)      # Sky emission
    
    
    # Emissivity calculations
    if refl is not None and emv is None:
        emv = 1. - refl # Assuming absorption is 1
    
    if rsoil is not None and ems is None:
        ems = 1. - rsoil
    
    if rsoil is None and ems is not None:
        rsoil = 1. - ems
    if refl is None and emv is not None:
        refl = 1. - emv
    
    [tss, too, tsstoo, rdd, tdd, rsd, tsd, rdo, tdo,
         rso, rsos, rsod, rddt, rsdt, rdot, rsodt, rsost, rsot,
         gammasdf, gammasdb, gammaso] = foursail (refl, np.zeros_like(refl),  
                                                  lidfa, lidfb, typelidf, 
                                                  lai, hspot, 
                                                  tts, tto, psi, rsoil)
    
    gammad = 1.0 - rdd - tdd
    gammao = 1.0 - rdo - tdo - too

    tso = tss*too+tss*(tdo+rsoil*rdd*too)/(1.0-rsoil*rdd)
    ttot = (too+tdo)/(1.0-rsoil*rdd)
    gammaot = gammao + ttot*rsoil*gammad
    gammasot = gammaso + ttot*rsoil*gammasdf

    aeev = gammaot
    aees = ttot*ems

    Lw = ( rdot*Hsky + 
            (aeev*Hc + 
            gammasot*emv*(Hh-Hc) + 
            aees*Hd + 
            tso*ems*(Hs-Hd)))/np.pi
    
    dnoem1 = top/(Lw*np.pi)
    Tbright = c2/(lam*np.log(dnoem1+1.0))
    dir_em = 1.0 - rdot 
    return Lw, Tbright, dir_em
        



#!/usr/bin/env python
from math import  exp, radians


try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache

import numpy as np
import numba


@numba.jit('Tuple((f8, f8, f8, f8))(f8,f8,f8,f8)',
           nopython=True, cache=True)
def volscatt(tts,tto,psi,ttl) :
    '''Compute volume scattering functions and interception coefficients
    for given solar zenith, viewing zenith, azimuth and leaf inclination angle.

    Parameters
    ----------
    tts : float
        Solar Zenith Angle (degrees).
    tto : float
        View Zenight Angle (degrees).
    psi : float
        View-Sun reliative azimuth angle (degrees).
    ttl : float
        leaf inclination angle (degrees).

    Returns
    -------
    chi_s : float
        Interception function  in the solar path.
    chi_o : float
        Interception function  in the view path.
    frho : float
        Function to be multiplied by leaf reflectance to obtain the volume scattering.
    ftau : float
        Function to be multiplied by leaf transmittance to obtain the volume scattering.

    References
    ----------
    Wout Verhoef, april 2001, for CROMA.
    '''


    cts = np.cos(np.radians(tts))
    cto = np.cos(np.radians(tto))
    sts = np.sin(np.radians(tts))
    sto = np.sin(np.radians(tto))
    cospsi = np.cos(np.radians(psi))
    psir = np.radians(psi)
    cttl = np.cos(np.radians(ttl))
    sttl = np.sin(np.radians(ttl))
    cs = cttl*cts
    co = cttl*cto
    ss = sttl*sts
    so = sttl*sto
    cosbts = 5.
    if np.abs(ss) > 1e-6 :
        cosbts = -cs/ss
    cosbto = 5.
    if np.abs(so) > 1e-6 :
        cosbto = -co/so
    if np.abs(cosbts) < 1.0:
        bts = np.arccos(cosbts)
        ds = ss
    else:
        bts = np.pi
        ds = cs
    chi_s = 2./np.pi*((bts-np.pi*0.5)*cs+np.sin(bts)*ss)
    if abs(cosbto) < 1.0:
        bto = np.arccos(cosbto)
        do_ = so
    else:
        if tto < 90.:
            bto = np.pi
            do_ = co
        else:
            bto = 0.0
            do_ = -co
    chi_o = 2.0/np.pi*((bto-np.pi*0.5)*co+np.sin(bto)*so)
    btran1 = np.abs(bts-bto)
    btran2 = np.pi-np.abs(bts+bto-np.pi)
    if psir <= btran1:
        bt1 = psir
        bt2 = btran1
        bt3 = btran2
    else:
        bt1 = btran1
        if psir <= btran2:
            bt2 = psir
            bt3 = btran2
        else:
            bt2 = btran2
            bt3 = psir
    t1 = 2.*cs*co+ss*so*cospsi
    t2 = 0.
    if bt2 > 0.:
        t2 = np.sin(bt2)*(2.*ds*do_+ss*so*np.cos(bt1)*np.cos(bt3))
    denom=2.*np.pi**2
    frho = ((np.pi-bt2)*t1+t2)/denom
    ftau = (-bt2*t1+t2)/denom
    if frho < 0. :
        frho=0.
    if ftau < 0. :
        ftau=0.

    return (chi_s,chi_o,frho,ftau)



@numba.jit('Tuple((f8, f8, f8, f8, f8))(f8[:], f8, f8, f8)',
           nopython=True, cache=True)
def weighted_sum_over_lidf (lidf, tts, tto, psi):
    ks = 0.
    ko = 0.
    bf = 0.
    sob = 0.
    sof = 0.
    cts   = np.cos(np.radians(tts))
    cto   = np.cos(np.radians(tto))
    ctscto  = cts*cto

    n_angles=len(lidf)
    angle_step=float(90.0/n_angles)
    litab = np.arange(n_angles)*angle_step + (angle_step*0.5)

    for i,ili in enumerate(litab):
        ttl = 1.*ili
        cttl=np.cos(np.radians(ttl))
        # SAIL volume scattering phase function gives interception and portions to be multiplied by rho and tau
        [chi_s,chi_o,frho,ftau]=volscatt(tts,tto,psi,ttl)
        # Extinction coefficients
        ksli=chi_s/cts
        koli=chi_o/cto
        # Area scattering coefficient fractions
        sobli=frho*np.pi/ctscto
        sofli=ftau*np.pi/ctscto
        bfli=cttl**2.
        ks+=ksli*float(lidf[i])
        ko+=koli*float(lidf[i])
        bf+=bfli*float(lidf[i])
        sob+=sobli*float(lidf[i])
        sof+=sofli*float(lidf[i])
    return ks, ko, bf, sob, sof



@lru_cache(maxsize=16)
def define_geometric_constants ( tts, tto, psi):
    cts   = np.cos(np.radians(tts))
    cto   = np.cos(np.radians(tto))
    ctscto  = cts*cto
    tants = np.tan(np.radians(tts))
    tanto = np.tan(np.radians(tto))
    cospsi  = np.cos(np.radians(psi))
    dso = np.sqrt(tants**2.+tanto**2.-2.*tants*tanto*cospsi)
    return cts, cto, ctscto, tants, tanto, cospsi, dso

@numba.jit("Tuple((f8,f8))(f8,f8,f8,f8)",
           nopython=True, cache=True)
def hotspot_calculations(alf, lai, ko, ks ):
    fhot=lai*np.sqrt(ko*ks)
    # Integrate by exponential Simpson method in 20 steps the steps are arranged according to equal partitioning of the slope of the joint probability function
    x1=0.
    y1=0.
    f1=1.
    fint=(1.-np.exp(-alf))*.05
    sumint=0.
    for istep in range(1,21):
        if istep < 20 :
            x2 = -np.log(1.-istep*fint)/alf
        else:
            x2=1.
        y2 = -(ko+ks)*lai*x2+fhot*(1.-np.exp(-alf*x2))/alf
        f2 = np.exp(y2)
        sumint = sumint+(f2-f1)*(x2-x1)/(y2-y1)
        x1 = x2
        y1 = y2
        f1 = f2
    tsstoo = f1
    if np.isnan(sumint) :
        sumint=0.
    return tsstoo, sumint

def Jfunc1(k,l,t) :
    ''' J1 function with avoidance of singularity problem.'''
    try:
        nb=len(l)
    except TypeError:
        nb = 1
    del_=(k-l)*t
    if nb > 1:
        result=np.zeros(nb)
        result[np.abs(del_) > 1e-3] = (np.exp(-l[np.abs(del_)> 1e-3]*t)-
                                  np.exp(-k*t))/(k-l[np.abs(del_)> 1e-3])
        result[np.abs(del_) <= 1e-3] = 0.5*t*(np.exp(-k*t) +
                                    np.exp(-l[np.abs(del_)<= 1e-3]*t))* \
                                    (1.-(del_[np.abs(del_)<= 1e-3]**2.)/12.)
    else:
        if np.abs(del_) > 1e-3 :
            result = (np.exp(-l*t)-np.exp(-k*t))/(k-l)
        else:
            result = 0.5*t*(np.exp(-k*t)+np.exp(-l*t))*(1.-(del_**2.)/12.)
    return result

def Jfunc2(k,l,t) :
    '''J2 function.'''
    return (1.-np.exp(-(k+l)*t))/(k+l)

@numba.jit('f8[:](f8,f8,i8)',
           nopython=True, cache=True)
def verhoef_bimodal(a,b,n_elements=18):
    '''Calculate the Leaf Inclination Distribution Function based on the
    Verhoef's bimodal LIDF distribution.
    Parameters
    ----------
    a : float
        controls the average leaf slope.
    b : float
        controls the distribution's bimodality.

            * LIDF type     [a,b].
            * Planophile    [1,0].
            * Erectophile   [-1,0].
            * Plagiophile   [0,-1].
            * Extremophile  [0,1].
            * Spherical     [-0.35,-0.15].
            * Uniform       [0,0].
            * requirement: |LIDFa| + |LIDFb| < 1.
    n_elements : int
        Total number of equally spaced inclination angles.

    Returns
    -------
    lidf : list
        Leaf Inclination Distribution Function at equally spaced angles.

    References
    ----------
    .. [Verhoef1998] Verhoef, Wout. Theory of radiative transfer models applied
        in optical remote sensing of vegetation canopies.
        Nationaal Lucht en Ruimtevaartlaboratorium, 1998.
        http://library.wur.nl/WebQuery/clc/945481.
        '''

    freq=1.0
    step=90.0/n_elements
    lidf=np.zeros(n_elements)*1.
    angles = (np.arange(n_elements)*step)[::-1]
    i = 0
    for angle in angles:
        tl1=np.radians(angle)
        if a>1.0:
            f = 1.0-np.cos(tl1)
        else:
            eps=1e-8
            delx=1.0
            x=2.0*tl1
            p=float(x)
            while delx >= eps:
                y = a*np.sin(x)+.5*b*np.sin(2.*x)
                dx=.5*(y-x+p)
                x=x+dx
                delx=abs(dx)
            f = (2.*y+p)/np.pi
        freq=freq-f
        lidf[i] = freq
        freq=float(f)
        i += 1
    lidf=lidf[::-1]
    return  lidf



@numba.jit('f8[:](f8,i8)',
           nopython=True, cache=True)
def campbell(alpha,n_elements=18):
    '''Calculate the Leaf Inclination Distribution Function based on the
    mean angle of [Campbell1990] ellipsoidal LIDF distribution.
    Parameters
    ----------
    alpha : float
        Mean leaf angle (degrees) use 57 for a spherical LIDF.
    n_elements : int
        Total number of equally spaced inclination angles .

    Returns
    -------
    lidf : list
        Leaf Inclination Distribution Function for 18 equally spaced angles.

    References
    ----------
    .. [Campbell1986] G.S. Campbell, Extinction coefficients for radiation in
        plant canopies calculated using an ellipsoidal inclination angle distribution,
        Agricultural and Forest Meteorology, Volume 36, Issue 4, 1986, Pages 317-321,
        ISSN 0168-1923, http://dx.doi.org/10.1016/0168-1923(86)90010-9.
    .. [Campbell1990] G.S Campbell, Derivation of an angle density function for
        canopies with ellipsoidal leaf angle distributions,
        Agricultural and Forest Meteorology, Volume 49, Issue 3, 1990, Pages 173-176,
        ISSN 0168-1923, http://dx.doi.org/10.1016/0168-1923(90)90030-A.
    '''


    alpha=float(alpha)
    excent=exp(-1.6184e-5*alpha**3.+2.1145e-3*alpha**2.-1.2390e-1*alpha+3.2491)
    sum0 = 0.
    freq = np.zeros(n_elements)
    step=90.0/n_elements
    for  i in range (n_elements):
        tl1=radians(i*step)
        tl2=radians((i+1.)*step)
        x1  = excent/(np.sqrt(1.+excent**2.*np.tan(tl1)**2.))
        x2  = excent/(np.sqrt(1.+excent**2.*np.tan(tl2)**2.))
        if excent == 1. :
            freq[i] = abs(np.cos(tl1)-np.cos(tl2))
        else :
            alph  = excent/np.sqrt(abs(1.-excent**2.))
            alph2 = alph**2.
            x12 = x1**2.
            x22 = x2**2.
            if excent > 1. :
                alpx1 = np.sqrt(alph2+x12)
                alpx2 = np.sqrt(alph2+x22)
                dum   = x1*alpx1+alph2*np.log(x1+alpx1)
                freq[i]=abs(dum-(x2*alpx2+alph2*np.log(x2+alpx2)))
            else :
                almx1 = np.sqrt(alph2-x12)
                almx2 = np.sqrt(alph2-x22)
                dum   = x1*almx1+alph2*np.arcsin(x1/alph)
                freq[i]=abs(dum-(x2*almx2+alph2*np.arcsin(x2/alph)))
    sum0 = np.sum(freq)
    lidf=np.zeros(n_elements)
    for i in range(n_elements):
        lidf[i] = freq[i]/sum0

    return lidf


def foursail (rho, tau, lidfa, lidfb, lidftype, lai, hotspot,
    tts, tto, psi, rsoil):
    """
    Parameters
    ----------
    rho : array_like
        leaf lambertian reflectance.
    tau : array_like
        leaf transmittance.
    lidfa : float
        Leaf Inclination Distribution at regular angle steps.
    lidfb : float
        Leaf Inclination Distribution at regular angle steps.
    lidftype : float
        Leaf Inclination Distribution at regular angle steps.
    lai : float
        Leaf Area Index.
    hotspot : float
        Hotspot parameter.
    tts : float
        Sun Zenith Angle (degrees).
    tto : float
        View(sensor) Zenith Angle (degrees).
    psi : float
        Relative Sensor-Sun Azimuth Angle (degrees).
    rsoil : array_like
        soil lambertian reflectance.

    Returns
    -------
    tss : array_like
        beam transmittance in the sun-target path.
    too : array_like
        beam transmittance in the target-view path.
    tsstoo : array_like
        beam tranmittance in the sur-target-view path.
    rdd : array_like
        canopy bihemisperical reflectance factor.
    tdd : array_like
        canopy bihemishperical transmittance factor.
    rsd : array_like
        canopy directional-hemispherical reflectance factor.
    tsd : array_like
        canopy directional-hemispherical transmittance factor.
    rdo : array_like
        canopy hemispherical-directional reflectance factor.
    tdo : array_like
        canopy hemispherical-directional transmittance factor.
    rso : array_like
        canopy bidirectional reflectance factor.
    rsos : array_like
        single scattering contribution to rso.
    rsod : array_like
        multiple scattering contribution to rso.
    rddt : array_like
        surface bihemispherical reflectance factor.
    rsdt : array_like
        surface directional-hemispherical reflectance factor.
    rdot : array_like
        surface hemispherical-directional reflectance factor.
    rsodt : array_like
        reflectance factor.
    rsost : array_like
        reflectance factor.
    rsot : array_like
        surface bidirectional reflectance factor.
    gammasdf : array_like
        'Thermal gamma factor'.
    gammasdb : array_like
        'Thermal gamma factor'.
    gammaso : array_like
        'Thermal gamma factor'.

    References
    ----------
    .. [Verhoef2007] Verhoef, W.; Jia, Li; Qing Xiao; Su, Z., (2007) Unified Optical-Thermal
        Four-Stream Radiative Transfer Theory for Homogeneous Vegetation Canopies,
        IEEE Transactions on Geoscience and Remote Sensing, vol.45, no.6, pp.1808-1822,
        http://dx.doi.org/10.1109/TGRS.2007.895844 based on  in Verhoef et al. (2007).
    """


    # Define some geometric constants.
    cts, cto, ctscto, tants, tanto, cospsi, dso = \
        define_geometric_constants(tts, tto, psi)

    # Calcualte leaf angle distribution
    if lidftype == 1:
        lidf = verhoef_bimodal(lidfa, lidfb, n_elements=18)
    elif lidftype == 2:
        lidf = campbell(lidfa, n_elements=18)
    else:
        raise ValueError(
            "lidftype can only be 1 (Campbell) or 2 (ellipsoidal)"
        )
    #Calculate geometric factors associated with extinction and scattering
    #Initialise sums
    ks=0.
    ko=0.
    bf=0.
    sob=0.
    sof=0.

    ks, ko, bf, sob, sof = weighted_sum_over_lidf(lidf, tts, tto, psi)


    # Geometric factors to be used later with rho and tau
    sdb = 0.5*(ks + bf)
    sdf = 0.5*(ks - bf)
    dob = 0.5*(ko + bf)
    dof = 0.5*(ko - bf)
    ddb = 0.5*(1.0 + bf)
    ddf = 0.5*(1.0 - bf)


    sigb=ddb*rho+ddf*tau
    sigf=ddf*rho+ddb*tau
    try:
        sigf[sigf == 0.0] = 1.e-36
        sigb[sigb == 0.0] = 1.0e-36
    except TypeError:
        sigf = max(1e-36, sigf)
        sigb = max(1e-36, sigb)
    att = 1.-sigf
    m = np.sqrt(att**2.- sigb**2.)
    sb = sdb*rho + sdf*tau
    sf = sdf*rho+sdb*tau
    vb = dob*rho+dof*tau
    vf = dof*rho+dob*tau
    w = sob*rho+sof*tau

    if lai<=0:
        # No canopy...
        tss = 1
        too= 1
        tsstoo= 1
        rdd= 0
        tdd=1
        rsd=0
        tsd=0
        rdo=0
        tdo=0
        rso=0
        rsos=0
        rsod=0
        rddt= rsoil
        rsdt= rsoil
        rdot= rsoil
        rsodt= 0
        rsost= rsoil
        rsot= rsoil
        gammasdf=0
        gammaso=0
        gammasdb=0

        return [tss,too,tsstoo,rdd,tdd,rsd,tsd,rdo,tdo,
            rso,rsos,rsod,rddt,rsdt,rdot,rsodt,rsost,rsot,gammasdf,gammasdb,gammaso]

    e1 = np.exp(-m*lai)
    e2 = e1**2.
    rinf = (att-m)/sigb
    rinf2 = rinf**2.
    re = rinf*e1
    denom = 1.-rinf2*e2
    J1ks = Jfunc1(ks,m,lai)
    J2ks = Jfunc2(ks,m,lai)
    J1ko = Jfunc1(ko,m,lai)
    J2ko = Jfunc2(ko,m,lai)
    Pss = (sf+sb*rinf)*J1ks
    Qss = (sf*rinf+sb)*J2ks
    Pv = (vf+vb*rinf)*J1ko
    Qv = (vf*rinf+vb)*J2ko
    tdd = (1.-rinf2)*e1/denom
    rdd = rinf*(1.-e2)/denom
    tsd = (Pss-re*Qss)/denom
    rsd = (Qss-re*Pss)/denom
    tdo = (Pv-re*Qv)/denom
    rdo = (Qv-re*Pv)/denom
    # Thermal "sd" quantities
    gammasdf = (1.+rinf)*(J1ks-re*J2ks)/denom
    gammasdb = (1.+rinf)*(-re*J1ks+J2ks)/denom
    tss = np.exp(-ks*lai)
    too = np.exp(-ko*lai)
    z = Jfunc2(ks,ko,lai)
    g1 = (z-J1ks*too)/(ko+m)
    g2 = (z-J1ko*tss)/(ks+m)
    Tv1 = (vf*rinf+vb)*g1
    Tv2 = (vf+vb*rinf)*g2
    T1 = Tv1*(sf+sb*rinf)
    T2 = Tv2*(sf*rinf+sb)
    T3 = (rdo*Qss+tdo*Pss)*rinf
    # Multiple scattering contribution to bidirectional canopy reflectance
    rsod = (T1+T2-T3)/(1.-rinf2)
    # Thermal "sod" quantity
    T4 = Tv1*(1.+rinf)
    T5 = Tv2*(1.+rinf)
    T6 = (rdo*J2ks+tdo*J1ks)*(1.+rinf)*rinf
    gammasod = (T4+T5-T6)/(1.-rinf2)
    #Treatment of the hotspot-effect
    alf=1e36
    # Apply correction 2/(K+k) suggested by F.-M. Breon
    if hotspot > 0. :
        alf=(dso/hotspot)*2./(ks+ko)
    if alf == 0. :
        # The pure hotspot
        tsstoo = tss
        sumint=(1.-tss)/(ks*lai)
    else :
        # Outside the hotspot
        tsstoo, sumint = hotspot_calculations(alf, lai, ko, ks)

    # Bidirectional reflectance
    # Single scattering contribution
    rsos=w*lai*sumint
    gammasos=ko*lai*sumint
    # Total canopy contribution
    rso=rsos+rsod
    gammaso=gammasos+gammasod
    #Interaction with the soil
    dn=1.-rsoil*rdd
    try:
        dn[ dn < 1e-36] = 1e-36
    except TypeError:
        dn = max ( 1e-36, dn)
    rddt=rdd+tdd*rsoil*tdd/dn
    rsdt=rsd+(tsd+tss)*rsoil*tdd/dn
    rdot=rdo+tdd*rsoil*(tdo+too)/dn
    rsodt=((tss+tsd)*tdo+(tsd+tss*rsoil*rdd)*too)*rsoil/dn
    rsost=rso+tsstoo*rsoil
    rsot=rsost+rsodt

    return [tss,too,tsstoo,rdd,tdd,rsd,tsd,rdo,tdo,
          rso,rsos,rsod,rddt,rsdt,rdot,rsodt,rsost,rsot,gammasdf,gammasdb,gammaso]




#!/usr/bin/env python
"""Spectral libraries for PROSPECT + SAIL
"""
import pkgutil
from collections import namedtuple
from io import BytesIO

import numpy as np

Spectra = namedtuple('Spectra', 'prospect5 prospectd soil light')
Prospect5Spectra = namedtuple('Prospect5Spectra', 
                                'nr kab kcar kbrown kw km')
ProspectDSpectra = namedtuple('ProspectDSpectra', 
                                'nr kab kcar kbrown kw km kant')
SoilSpectra = namedtuple("SoilSpectra", "rsoil1 rsoil2")
LightSpectra = namedtuple("LightSpectra", "es ed")

def get_spectra():
    """Reads the spectral information and stores is for future use."""

    # PROSPECT-D
    prospect_d_spectraf = pkgutil.get_data('prosail', 'prospect_d_spectra.txt')
    _, nr, kab, kcar, kant, kbrown, kw, km= np.loadtxt(
        BytesIO(prospect_d_spectraf), unpack=True)
    prospect_d_spectra = ProspectDSpectra(nr, kab, kcar, kbrown, kw, km, kant)
    # PROSPECT 5
    prospect_5_spectraf = pkgutil.get_data('prosail', 'prospect5_spectra.txt')
    nr, kab, kcar, kbrown, kw, km =  np.loadtxt(BytesIO(prospect_5_spectraf),
                                                unpack=True)
    prospect_5_spectra = Prospect5Spectra(nr, kab, kcar, kbrown, kw, km)
    # SOIL
    soil_spectraf = pkgutil.get_data('prosail', 'soil_reflectance.txt')
    rsoil1, rsoil2 =  np.loadtxt(BytesIO(soil_spectraf),
                                                unpack=True)
    soil_spectra = SoilSpectra(rsoil1, rsoil2)    
    # LIGHT
    light_spectraf = pkgutil.get_data('prosail', 'light_spectra.txt')
    es, ed =  np.loadtxt(BytesIO(light_spectraf),
                                                unpack=True)
    light_spectra = LightSpectra(es, ed)
    spectra = Spectra(prospect_5_spectra, prospect_d_spectra, 
                      soil_spectra, light_spectra)
    return spectra
